{# Version switcher flyout for GitHub Pages hosted docs #}
{% if versions %}
<div class="rst-versions" data-toggle="rst-versions" role="note" aria-label="Versions">
  <span class="rst-current-version" data-toggle="rst-current-version">
    <span class="fa fa-book"></span> Version:
    <span class="version-name">{{ current_version }}</span>
    <span class="fa fa-caret-down"></span>
  </span>
  <div class="rst-other-versions">
    <dl>
      <dt>Versions</dt>
      {% for name, url in versions %}
        <dd><a href="{{ url }}">{{ name }}</a></dd>
      {% endfor %}
    </dl>
  </div>
</div>
{% endif %}
