:root {
  --color-a: #7f1e79;
  --color-b: #7f1e48;
  --color-c: #551e7f;
  --color-a-light: #7f1e7933;
  --color-a-bright: #c46bbe;
  --color-b-light: #7f1e4833;
  --color-c-light: #551e7f33;
}

.wy-nav-top {
  background: var(--color-a);
}

.wy-side-nav-search {
  background: var(--color-a);
}

a {
  color: var(--color-a);
}
a:hover {
  color: var(--color-c);
}

.wy-menu-vertical a {
  filter: brightness(1.8);
}

.wy-menu-vertical a:active {
  background-color: var(--color-b);
}

.wy-menu-vertical a:visited {
  filter: brightness(1.8);
}

.rst-content .admonition-title {
  background: var(--color-a);
}

.rst-content .hint .admonition-title {
  background: var(--color-a-light);
}

button.toggle-button {
  color: var(--color-a-light);
}

.rst-content .admonition {
  background: var(--color-a-light);
}

.rst-content .admonition-title::before {
  content: "";
}

.rst-content code.literal {
  color: var(--color-a);
}

html.writer-html4 .rst-content dl:not(.docutils) > dt {
  color: var(--color-a);
  border-top-color: var(--color-a);
  background: var(--color-a-light);
}
html.writer-html5 .rst-content dl[class]:not(.option-list):not(.field-list):not(.footnote):not(.glossary):not(.simple) > dt {
  color: var(--color-a);
  border-top-color: var(--color-a);
  background: var(--color-a-light);
}

.rst-versions {
  background: #1f1d53;
  font-family: "Lato", "proxima-nova", "Helvetica Neue", Arial, sans-serif;
}

.rst-versions .rst-current-version {
  color: white;
  display: flex;
  align-items: center;
}

.rst-versions .rst-current-version .fa-book {
  margin-right: 0.4em;
}

.rst-versions .rst-current-version .version-name {
  color: var(--color-a-bright);
  margin-left: auto;
}

.rst-versions .rst-current-version .fa-caret-down {
  margin-left: 0.4em;
}

.rst-versions a,
.rst-versions .rst-other-versions dd a {
  color: var(--color-a-bright);
}

.rst-versions .rst-other-versions dd a:hover {
  color: white;
}
